/* Dark mode */
[data-bs-theme="dark"] .Select-control {
    background-color: rgb(25, 25, 25) !important;
    color: white;
}

[data-bs-theme="dark"] .Select-menu-outer {
    background-color: rgb(25, 25, 25);
    color: white;
}

[data-bs-theme="dark"] .Select-value-label {
    color: white !important;
}

/* Light mode */
[data-bs-theme="light"] .Select-control {
    background-color: white !important;
    color: black;
}

[data-bs-theme="light"] .Select-menu-outer {
    background-color: white;
    color: black;
}

[data-bs-theme="light"] .Select-value-label {
    color: black !important;
}